#!/usr/bin/env python3
import numpy as np
from skimage import measure, morphology, transform
from scipy.interpolate import RegularGridInterpolator
import scipy as sp
from scipy.stats import beta, mode

from matplotlib.colors import LogNorm
import seaborn as sns
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import Size, Divider
from imageio import imsave


def nfa_from_block_votes(grids, down):
    grids = grids.copy()
    min_size = down + 1
    disk_size = 2
    disk = morphology.disk(disk_size)
    step = max(int(np.sqrt(down)), 16)
    diag = -np.ones_like(grids)
    diag[grids >= 0] = grids[grids >= 0] // 2
    diag[grids < -1] = -grids[grids < -1] - 2
    grids[grids < -1] = -1

    diag_global = int(mode(diag[diag >= 0]).mode[0])
    grid_global = int(mode(grids[grids >= 0]).mode[0])

    is_incorrect_diag = (diag != diag_global) * (diag >= 0)
    is_incorrect_grid = (grids != grid_global) * (grids >= 0)

    p0_grid = np.mean(is_incorrect_grid[grids >= 0])
    p0_diag = np.mean(is_incorrect_diag[diag >= 0])

    is_incorrect = is_incorrect_diag
    is_valid = diag >= 0
    p0 = p0_diag
    nfa_inconsistent_diag = np.ones(is_incorrect.shape) * np.inf
    nfa_inconsistent = nfa_inconsistent_diag
    map_ = sp.ndimage.gaussian_filter(is_incorrect.astype(float), 7) > 1.1 * p0
    map_ = morphology.binary_closing(map_, disk)
    labels, i_regions = measure.label(map_, return_num=True)

    for i_component in range(1, i_regions + 1):
        mask = labels == i_component
        yy, xx = np.argwhere(mask).T
        Y_min, Y_max = max(0,
                           yy.min() - disk_size), min(yy.max() + disk_size,
                                                      mask.shape[0])
        X_min, X_max = max(0,
                           xx.min() - disk_size), min(xx.max() + disk_size,
                                                      mask.shape[1])
        if (Y_max - Y_min) * (X_max - X_min) < min_size:
            continue
        step_y = max(1, min(step, (Y_max - Y_min) // 4))
        step_x = max(1, min(step, (X_max - X_min) // 4))
        mesh = np.mgrid[Y_min:Y_max:step_y, Y_min:Y_max:step_y,
                        X_min:X_max:step_x, X_min:X_max:step_x]
        ys, ye, xs, xe = mesh
        is_large_enough = (ye > ys) * (xe > xs) * ((xe - xs) *
                                                   (ye - ys) > min_size)
        if not np.any(is_large_enough):
            continue
        mesh = mesh[:, is_large_enough]
        ys, ye, xs, xe = mesh
        ys = ys - Y_min
        ye = ye - Y_min
        xs = xs - X_min
        xe = xe - X_min
        small_is_incorrect = is_incorrect[Y_min:Y_max, X_min:X_max]
        ii = transform.integral.integral_image(small_is_incorrect)
        counts = transform.integral.integrate(ii,
                                              np.stack((ys, xs)).T,
                                              np.stack((ye - 1, xe - 1)).T)
        #totals = (ye - ys) * (xe - xs)
        totals = [
            np.count_nonzero(is_valid[yys:yye, xxs:xxe])
            for (yys, yye, xxs, xxe) in zip(ys, ye, xs, xe)
        ]
        a = counts / down - 1
        b = (totals - counts) / down
        a = np.maximum(a, 1e-20)
        b = np.maximum(b, 1e-20)
        nfas = diag.size / down * beta.cdf(p0, a, b)
        nfa = nfas.min()
        mask = labels == i_component
        mask = morphology.binary_closing(mask, morphology.disk(8))
        nfa_inconsistent[mask] = np.minimum(nfa_inconsistent[mask], nfa)

        is_incorrect = is_incorrect_grid

    is_valid = grids >= 0
    p0 = p0_grid
    nfa_inconsistent_grid = np.ones(is_incorrect.shape) * np.inf
    nfa_inconsistent = nfa_inconsistent_grid
    map_ = sp.ndimage.gaussian_filter(is_incorrect.astype(float), 7) > 1.1 * p0
    map_ = morphology.binary_closing(map_, disk)
    labels, i_regions = measure.label(map_, return_num=True)
    for i_component in range(1, i_regions + 1):
        mask = labels == i_component
        yy, xx = np.argwhere(mask).T
        Y_min, Y_max = max(0,
                           yy.min() - disk_size), min(yy.max() + disk_size,
                                                      mask.shape[0])
        X_min, X_max = max(0,
                           xx.min() - disk_size), min(xx.max() + disk_size,
                                                      mask.shape[1])
        if (Y_max - Y_min) * (X_max - X_min) < min_size:
            continue
        step_y = max(1, min(step, (Y_max - Y_min) // 4))
        step_x = max(1, min(step, (X_max - X_min) // 4))
        mesh = np.mgrid[Y_min:Y_max:step_y, Y_min:Y_max:step_y,
                        X_min:X_max:step_x, X_min:X_max:step_x]
        ys, ye, xs, xe = mesh
        is_large_enough = (ye > ys) * (xe > xs) * ((xe - xs) *
                                                   (ye - ys) > min_size)
        if not np.any(is_large_enough):
            continue
        mesh = mesh[:, is_large_enough]
        ys, ye, xs, xe = mesh
        ys = ys - Y_min
        ye = ye - Y_min
        xs = xs - X_min
        xe = xe - X_min
        small_is_incorrect = is_incorrect[Y_min:Y_max, X_min:X_max]
        ii = transform.integral.integral_image(small_is_incorrect)
        counts = transform.integral.integrate(ii,
                                              np.stack((ys, xs)).T,
                                              np.stack((ye - 1, xe - 1)).T)
        #totals = (ye - ys) * (xe - xs)
        totals = [
            np.count_nonzero(is_valid[yys:yye, xxs:xxe])
            for (yys, yye, xxs, xxe) in zip(ys, ye, xs, xe)
        ]
        a = counts / down - 1
        b = (totals - counts) / down
        a = np.maximum(a, 1e-20)
        b = np.maximum(b, 1e-20)
        nfas = diag.size / down * beta.cdf(p0, a, b)
        nfa = nfas.min()
        mask = labels == i_component
        mask = morphology.binary_closing(mask, morphology.disk(8))
        nfa_inconsistent[mask] = np.minimum(nfa_inconsistent[mask], nfa)
    return nfa_inconsistent_diag, nfa_inconsistent_grid


def create_fig(figsize):
    """Creates a figure, axis and colorbar axis such that the main axis has the specified size after saving."""
    px = 1 / plt.rcParams['figure.dpi']
    sY, sX = figsize

    pad = sX * 0.05
    cbsz = sX * 0.05
    fig = plt.figure(figsize=((sX + pad + cbsz) * px, sY * px))

    # The first items are for padding and the second items are for the axes.
    # sizes are in inch.
    h = [Size.Fixed(0 * px), Size.Fixed(sX * px)]
    v = [Size.Fixed(0 * px), Size.Fixed(sY * px)]
    divider = Divider(fig, (0, 0, 1 - (pad + cbsz) / (sY + pad + cbsz), 1),
                      h,
                      v,
                      aspect=False)
    # The width and height of the rectangle are ignored.
    ax = fig.add_axes(divider.get_position(),
                      axes_locator=divider.new_locator(nx=1, ny=1))
    h = [Size.Fixed(0 * px), Size.Fixed(cbsz * px)]
    v = [Size.Fixed(0 * px), Size.Fixed(sY * px)]
    divider = Divider(fig, (1 - (pad + cbsz) / (sX + pad + cbsz), 0, 1, 1),
                      h,
                      v,
                      aspect=False)
    cax = fig.add_axes(divider.get_position(),
                       axes_locator=divider.new_locator(nx=1, ny=1))
    return fig, ax, cax


def interpolate_nfa(grids, border, target_shape):
    cuts_y = np.arange(grids.shape[0]) * 2 + 1 + border
    cuts_x = np.arange(grids.shape[1]) * 2 + 1 + border
    interpolator = RegularGridInterpolator((cuts_y, cuts_x),
                                           grids,
                                           method='nearest',
                                           bounds_error=False,
                                           fill_value=np.inf)
    target_coords = np.asarray(
        np.meshgrid(np.arange(target_shape[0]), np.arange(target_shape[1])))
    result = interpolator(target_coords.reshape(2, -1).T).reshape(
        target_shape[1], target_shape[0]).T
    return result


def save_nfa(out, nfa, figsize):
    fig, ax, cax = create_fig(figsize)
    plot = ax.matshow(np.clip(nfa, 1e-12, 1e6), norm=LogNorm(), cmap='mako_r')
    plt.colorbar(plot, cax=cax, extend='both')
    cax.tick_params(labelsize=35)
    ax.axis('off')
    plt.savefig(out, bbox_inches='tight', pad_inches=0, transparent=False)


if __name__ == "__main__":
    from argparse import ArgumentParser
    from sys import argv
    parser = ArgumentParser()
    parser.add_argument('input', type=str)
    parser.add_argument('output', type=str)
    parser.add_argument('output2', type=str)
    args = parser.parse_args(argv[1:])
    data = np.load(args.input)
    grids = data['grids']
    nfa_diag, nfa_grid = nfa_from_block_votes(data['grids'], data['down'])
    nfa = np.minimum(nfa_diag, nfa_grid)
    nfa = interpolate_nfa(nfa, data['border'], data['target_shape'])
    save_nfa(args.output, nfa, nfa.shape)
    imsave(args.output2, (nfa < 1e-3).astype(np.uint8) * 255)
